from dependency_metrics.constants import PIP, YARN
from dependency_metrics.package_managers.pip import get_pip_packages
from dependency_metrics.package_managers.yarn import get_yarn_packages
from dependency_metrics.parsing_utils import behind


def iter_packages(package_manager):
    """
    Obtains packages given a manager, and builds package info tuple
    :param package_manager: str representing package manager
    :return: generator of package info (behind, name, latest, current)
    """
    for package_info in get_packages(package_manager):
        latest = package_info["latest_version"]
        current = package_info["version"]
        yield behind(latest, current), package_info["name"], latest, current


def get_packages(package_manager):
    """
    :param package_manager: str representing package manager
    :return: function to retrieve packages for the provided package manager
    """
    package_list_map = {
        PIP: get_pip_packages,
        YARN: get_yarn_packages,
    }

    return package_list_map[package_manager]()



DATADOG_API_KEY = 'DATADOG_API_KEY'
DATADOG_APP_KEY = 'DATADOG_APP_KEY'
REPO = "REPO"

PIP = 'pip'
YARN = 'yarn'

UNKNOWN_VERSION = "unknown"


class MetricType:
    GAUGE = 'gauge'
    COUNT = 'count'
    RATE = 'rate'

    @classmethod
    def options(cls):
        return [cls.GAUGE, cls.COUNT, cls.RATE]



import json
import os
from unittest import TestCase
from unittest.mock import patch

from freezegun import freeze_time

from dependency_metrics.constants import MetricType
from dependency_metrics.datadog_utils import send_metric, send_stats_to_datadog


@patch('dependency_metrics.datadog_utils.send_metric')
def test_send_stats_to_datadog(mock_send_metric):
    stats = {'Outdated': 15}
    send_stats_to_datadog(stats, 'pip')
    mock_send_metric.assert_called_with(
        "commcare.static_analysis.dependency.python.outdated",
        15,
        MetricType.GAUGE)


@patch('dependency_metrics.datadog_utils.requests.post')
class SendMetricTests(TestCase):

    def test_raises_error_if_api_key_not_found(self, mock_post):
        self._setup_env(DATADOG_APP_KEY='app_key')

        with self.assertRaises(KeyError):
            send_metric('name', 'value', MetricType.GAUGE)

    def test_raises_error_if_app_key_not_found(self, mock_post):
        self._setup_env(DATADOG_API_KEY='api_key')

        with self.assertRaises(KeyError):
            send_metric('name', 'value', MetricType.GAUGE)

    def test_raises_error_if_metric_type_not_valid(self, mock_post):
        self._setup_env(DATADOG_API_KEY='api_key', DATADOG_APP_KEY='app_key')

        with self.assertRaises(ValueError):
            send_metric('name', 'value', 'invalid')

    @freeze_time('2020-01-01 10:20:30')
    def test_default_values(self, mock_post):
        self._setup_env(DATADOG_API_KEY='api_key', DATADOG_APP_KEY='app_key')

        send_metric('name', 'value', MetricType.GAUGE)

        args, kwargs = mock_post.call_args
        payload = kwargs['json']
        self.assertEqual(payload['series'][0]['metric'], 'name')
        self.assertEqual(payload['series'][0]['points'], [[1577874030, 'value']])
        self.assertEqual(payload['series'][0]['host'], 'unknown')

    def test_repository_set_as_host(self, mock_post):
        self._setup_env(DATADOG_API_KEY='api_key', DATADOG_APP_KEY='app_key',
                        REPO='test-repo')

        send_metric('name', 'value', MetricType.GAUGE)

        args, kwargs = mock_post.call_args
        payload = kwargs['json']
        self.assertEqual(payload['series'][0]['host'], 'test-repo')

    def test_metric_type_is_gauge_when_set(self, mock_post):
        self._setup_env(DATADOG_API_KEY='api_key', DATADOG_APP_KEY='app_key')

        send_metric('name', 'value', MetricType.GAUGE)

        args, kwargs = mock_post.call_args
        payload = kwargs['json']
        self.assertEqual(payload['series'][0]['type'], MetricType.GAUGE)

    def test_metric_type_is_count_when_set(self, mock_post):
        self._setup_env(DATADOG_API_KEY='api_key', DATADOG_APP_KEY='app_key')

        send_metric('name', 'value', MetricType.COUNT)

        args, kwargs = mock_post.call_args
        payload = kwargs['json']
        self.assertEqual(payload['series'][0]['type'], MetricType.COUNT)

    def test_metric_type_is_rate_when_set(self, mock_post):
        self._setup_env(DATADOG_API_KEY='api_key', DATADOG_APP_KEY='app_key')

        send_metric('name', 'value', MetricType.RATE)

        args, kwargs = mock_post.call_args
        payload = kwargs['json']
        self.assertEqual(payload['series'][0]['type'], MetricType.RATE)

    def _setup_env(self, **kwargs):
        """
        Useful to simplify test setup
        """
        for key, value in kwargs.items():
            os.environ[key] = value
            self.addCleanup(os.environ.pop, key)



from unittest import TestCase

from dependency_metrics.parsing_utils import behind, parse_version


class BehindTests(TestCase):
    """
    Tests ``behind(latest, current)``
    """

    def test_major_delta(self):
        delta = behind("2.0.0", "1.0.0")
        self.assertEqual(delta, [1, 0, 0])

    def test_minor_delta(self):
        delta = behind("1.1.0", "1.0.0")
        self.assertEqual(delta, [0, 1, 0])

    def test_patch_delta(self):
        delta = behind("1.0.1", "1.0.0")
        self.assertEqual(delta, [0, 0, 1])

    def test_major_has_priority_over_minor(self):
        delta = behind("2.1.0", "1.0.0")
        self.assertEqual(delta, [1, 0, 0])

    def test_major_has_priority_over_patch(self):
        delta = behind("2.0.1", "1.0.0")
        self.assertEqual(delta, [1, 0, 0])

    def test_minor_has_priority_over_patch(self):
        delta = behind("1.1.1", "1.0.0")
        self.assertEqual(delta, [0, 1, 0])

    def test_multi_major_delta(self):
        delta = behind("5.0.0", "1.0.0")
        self.assertEqual(delta, [4, 0, 0])

    def test_multi_minor_delta(self):
        delta = behind("1.5.0", "1.0.0")
        self.assertEqual(delta, [0, 5, 0])

    def test_multi_patch_delta(self):
        delta = behind("1.0.5", "1.0.0")
        self.assertEqual(delta, [0, 0, 5])

    def test_unknown_latest_version(self):
        delta = behind("unknown", "1.0.0")
        self.assertEqual(delta, None)


class ParseVersionTests(TestCase):

    def test_major_minor_patch(self):
        version = parse_version("3.2.1")
        self.assertEqual(version, [3, 2, 1])

    def test_major_minor(self):
        version = parse_version("3.2")
        self.assertEqual(version, [3, 2, 0])

    def test_major(self):
        version = parse_version("3")
        self.assertEqual(version, [3, 0, 0])

    def test_non_semvar(self):
        version = parse_version("4.3.2.1")
        self.assertEqual(version, [4, 3, 2])

    def test_named_version(self):
        version = parse_version("alpha")
        self.assertEqual(version, [0, 0, 0])

    def test_combination_of_numbers_and_names(self):
        version = parse_version("2.1rc1")
        self.assertEqual(version, [2, 1, 0])

    def test_multi_digit_number_in_name(self):
        version = parse_version("2.10rc1")
        self.assertEqual(version, [2, 10, 0])

    def test_number_after_name_not_used(self):
        version = parse_version("3.a3.1")
        self.assertEqual(version, [3, 0, 1])



import re


def behind(latest, current):
    if latest == 'unknown':
        return None
    delta = [0, 0, 0]
    parsed_latest = parse_version(latest)
    parsed_current = parse_version(current)
    for index, (lat, cur) in enumerate(zip(parsed_latest, parsed_current)):
        if lat != cur:
            delta[index] = lat - cur
            break
    return delta


def parse_version(version):
    """
    :param version: a string representing the package version (e.g., '1.0.1')
    :return: a list of integers representing [major, minor, patch]
    """
    padded = f"{version}.0.0"  # append extra zeros for versions like '1' or '2.0'
    semvar_version = []
    for n in padded.split(".")[:3]:
        try:
            num = int(n)
        except ValueError:
            match = re.search(r'^\d+', n)
            # default to 0, unless
            num = int(match.group()) if match else 0
        semvar_version.append(num)

    return semvar_version




import json
from json import JSONDecodeError

import sh

from dependency_metrics.constants import UNKNOWN_VERSION
from dependency_metrics.exceptions import Crash

"""
Yarn doesn't have any explicitly stated naming conventions for packages.
See here: https://yarnpkg.com/configuration/manifest. It is unclear if 
package names are case-sensitive or not.

When using ``yarn info`` for a specific package, yarn expects the name defined
in package.json to be used, which is not necessarily the same as the name 
fetched via ``yarn list``.

Example:
``yarn list --depth 0 | grep At.js`` => At.js@1.5.3
``yarn info At.js dist-tags.latest --json`` => {"type": "error", "data": "Received invalid response from npm."}
``yarn info at.js dist-tags.latest --json`` => {"type":"inspect","data":"1.5.4"}

The current behavior is to only use the name obtained from ``yarn list``. If 
that results in an error, the code below sets ``latest_version`` to "unknown". 
This has been the behavior up until now, so this will continue to be the 
behavior.

If new information is found that clearly states yarn package names are 
case-insensitive, then the behavior can change to retry package names that fail 
with a lower cased version of the name.
"""


def get_yarn_packages():
    version = Yarn.version()
    if not version.startswith("1."):
        raise Crash(f"Yarn Classic (v1.x) is required, found {version}")

    outdated_packages = []
    for package in parse_yarn_list():
        latest_version = pull_latest_version(package["name"])
        if package["version"] != latest_version:
            # only care if it is actually outdated
            package["latest_version"] = latest_version
            outdated_packages.append(package)
    return outdated_packages


def pull_latest_version(package_name):
    """
    Attempts to pull latest version of package from yarn
    :param package_name: name of javascript package
    :returns: returns version as string, or UNKNOWN_VERSION if not found
    """
    string_output = Yarn.latest_version(package_name)
    if not string_output:
        return UNKNOWN_VERSION

    try:
        json_output = json.loads(string_output)
    except JSONDecodeError:
        return UNKNOWN_VERSION

    if json_output.get('type') == 'inspect':
        return json_output.get('data')

    return UNKNOWN_VERSION


def parse_yarn_list():
    """
    Parse output of ``Yarn.list`` into list of package name and version dicts
    """
    packages = []
    yarn_list = Yarn.list()
    # trim first and last line from output as they are irrelevant
    for line in yarn_list.splitlines()[1:-1]:
        line = line.rstrip()
        if not line:
            continue
        # example:
        # ├─ @ungap/promise-all-settled@1.1.2
        # ├─ abbrev@1.1.1
        name, x, version = line.split(None, 2)[1].rpartition("@")
        packages.append({"name": name, "version": version})
    return packages


class Yarn:
    """
    Class to house raw yarn commands
    """

    @staticmethod
    def latest_version(package_name):
        """
        Equivalent to ``yarn info <package_name> dist-tags.latest --json``
        """
        return sh.yarn("info", package_name, "dist-tags.latest", "--json")

    @staticmethod
    def list():
        """
        Equivalent to ``yarn list --depth 0``
        """
        return sh.yarn("list", "--depth", "0")

    @staticmethod
    def version():
        """
        Equivalent to ``yarn --version``
        """
        return sh.yarn("--version")



import os
import time

import requests

from dependency_metrics.constants import (
    DATADOG_API_KEY,
    DATADOG_APP_KEY,
    PIP,
    REPO,
    YARN,
    MetricType,
)


def send_stats_to_datadog(stats, package_manager):
    """
    :param stats: key-value pairs representing outdated counts
    Example: {'major': 3, 'minor': 2, ...})
    :param package_manager: package manager used to collect statistics
    """
    metric_env = get_metric_name_for_package_manager(package_manager)
    metric_prefix = f"commcare.static_analysis.dependency.{metric_env}"
    for key, value in stats.items():
        metric_name = get_metric_name_for_stats_key(key)
        send_metric(f"{metric_prefix}.{metric_name}", value, MetricType.GAUGE)


def send_metric(name, value, metric_type, tags=None):
    """
    Send a specific metric to datadog
    :param name: name of metric
    :param value:
    :param metric_type: str in constants.MetricType
    :param tags: optional {key: value}, append additional tags to metric
           e.g., {"tag_name": "tag_value"}
    """

    api_key = os.environ[DATADOG_API_KEY]
    app_key = os.environ[DATADOG_APP_KEY]

    if metric_type not in MetricType.options():
        raise ValueError(
            f"Expected {metric_type} to be one of the following: "
            f"{MetricType.options()}")

    current_time = int(time.time())  # drop milliseconds from time.time()

    url = 'https://app.datadoghq.com/api/v1/series'
    headers = {
        'Content-type': 'application/json',
        'DD-API-KEY': api_key,
        'DD-APP-KEY': app_key,
    }

    tags = tags or {}
    payload = {
        'series': [{
            'metric': name,
            'points': [[current_time, value]],
            'type': metric_type,
            'host': os.environ.get(REPO) or 'unknown',
            'tags': [f'{key}:{value}' for key, value in tags.items()]
        }]
    }
    response = requests.post(url, headers=headers, json=payload)
    if not response.ok:
        print(
            f"Post request to {url} failed with status code "
            f"{response.status_code} and error {response.text}\n"
            f"The payload sent was {payload}"
        )


def get_metric_name_for_package_manager(key):
    metric_name_map = {
        PIP: "python",
        YARN: "js",
    }
    return metric_name_map[key]


def get_metric_name_for_stats_key(key):
    metric_name_map = {
        "Outdated": "outdated",
        "Multi-Major": "multi_major_outdated",
        "Major": "major_outdated",
        "Minor": "minor_outdated",
        "Patch": "patch_outdated",
        "Unknown": "unknown"
    }
    return metric_name_map[key]




import json

import sh


def get_pip_packages():
    """
    Return relevant package version info from pip list output
    """
    pip_packages = json.loads(Pip.list())
    cleaned_package_info_list = []
    for pkg_info in pip_packages:
        del pkg_info['latest_filetype']
        cleaned_package_info_list.append(pkg_info)
    return cleaned_package_info_list


class Pip:

    @staticmethod
    def list():
        """
        Equivalent to ``pip list --format json --outdated``
        """
        return sh.pip("list", "--format", "json", "--outdated")



import argparse

from dependency_metrics.constants import PIP, YARN
from dependency_metrics.datadog_utils import send_stats_to_datadog
from dependency_metrics.package_managers.utils import iter_packages


def build_packages_table(packages):
    """ Builds a table representing how out of date dependencies are
    Example:
    Behind   Package                      Latest       Version
    n/a      At.js                        exotic       1.5.3
    ...
    0.0.1    bootstrap-timepicker         0.5.2        0.5.1
    ...
    12.0.0   sinon                        14.0.0       2.3.2
    """
    def build_row(behind, name, current, latest):
        return f"{behind:8s} {name:28s} {current:12s} {latest}"

    rows = [build_row("Behind", "Package", "Latest", "Version")]

    records = sorted(packages)
    for delta, name, current, latest in records:
        if delta:
            behind = ".".join(str(v) for v in delta)
        else:
            behind = "n/a"
        rows.append(build_row(behind, name, current, latest))

    return rows


def get_package_stats(packages):
    """Displays a count for each version category of out of date dependencies"""
    stats = {
        "Outdated": 0,
        "Multi-Major": 0,
        "Major": 0,
        "Minor": 0,
        "Patch": 0,
        "Unknown": 0,
    }
    for delta, name, current, latest in packages:
        if delta:
            major, minor, patch = delta
            if major:
                assert not minor and not patch, delta
                if major == 1:
                    key = "Major"
                else:
                    key = "Multi-Major"
            elif minor:
                assert not major and not patch, delta
                key = "Minor"
            else:
                assert patch and not major and not minor, delta
                key = "Patch"
        else:
            key = "Unknown"
        stats[key] += 1
        stats["Outdated"] += 1
    return stats


def main():
    parser = argparse.ArgumentParser(
        description="Track dependencies of a project"
    )
    parser.add_argument(
        "package_manager",
        help="package manager to calculate metrics for",
        choices=[PIP, YARN]
    )
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument(
        "--stats",
        help="generate statistics and output to stdout",
        default=False,
        action="store_true",
    )
    group.add_argument(
        "--send",
        help="generate statistics and send to datadog",
        default=False,
        action="store_true"
    )
    args = parser.parse_args()

    packages = iter_packages(args.package_manager)
    if args.stats or args.send:
        stats = get_package_stats(packages)
        if args.stats:
            # NOTE: subtle detail: we're depending on Python 3's ordered dict to
            # maintain deterministic ordering here
            for key, value in stats.items():
                print(f"{key}: {value}")
        elif args.send:
            send_stats_to_datadog(stats, args.package_manager)

    else:
        package_table = build_packages_table(packages)
        for row in package_table:
            print(row)


if __name__ == "__main__":
    main()



from unittest import TestCase

from dependency_metrics.metrics import build_packages_table, get_package_stats


class BuildPackagesTableTests(TestCase):

    def test_header_row(self):
        packages = []
        rows = build_packages_table(packages)
        self.assertEqual(rows[0], "Behind   Package                      Latest       Version")

    def test_row_with_delta(self):
        packages = [([0, 1, 0], 'test', '1.1', '1.0')]
        rows = build_packages_table(packages)
        self.assertEqual(rows[1], "0.1.0    test                         1.1          1.0")

    def test_row_with_no_delta(self):
        packages = [([0, 0, 0], 'test', '1.0', '1.0')]
        rows = build_packages_table(packages)
        self.assertEqual(rows[1], "0.0.0    test                         1.0          1.0")


class GetPackageStatsTests(TestCase):

    def test_outdated_multi_major_package(self):
        packages = [([2, 0, 0], 'test', '3.1', '1.0')]
        stats = get_package_stats(packages)
        self.assertEqual(stats, {
            "Outdated": 1,
            "Multi-Major": 1,
            "Major": 0,
            "Minor": 0,
            "Patch": 0,
            "Unknown": 0,
        })

    def test_outdated_major_package(self):
        packages = [([1, 0, 0], 'test', '2.5', '1.0')]
        stats = get_package_stats(packages)
        self.assertEqual(stats, {
            "Outdated": 1,
            "Multi-Major": 0,
            "Major": 1,
            "Minor": 0,
            "Patch": 0,
            "Unknown": 0,
        })

    def test_outdated_minor_package(self):
        packages = [([0, 5, 0], 'test', '2.5', '2.0')]
        stats = get_package_stats(packages)
        self.assertEqual(stats, {
            "Outdated": 1,
            "Multi-Major": 0,
            "Major": 0,
            "Minor": 1,
            "Patch": 0,
            "Unknown": 0,
        })

    def test_outdated_patch_package(self):
        packages = [([0, 0, 3], 'test', '2.5.4', '2.5.1')]
        stats = get_package_stats(packages)
        self.assertEqual(stats, {
            "Outdated": 1,
            "Multi-Major": 0,
            "Major": 0,
            "Minor": 0,
            "Patch": 1,
            "Unknown": 0,
        })

    def test_unknown_package(self):
        packages = [(None, 'test', 'unknown', '2.5.1')]
        stats = get_package_stats(packages)
        self.assertEqual(stats, {
            "Outdated": 1,
            "Multi-Major": 0,
            "Major": 0,
            "Minor": 0,
            "Patch": 0,
            "Unknown": 1,
        })
